/*

*/
import java.awt.Rectangle;
import java.util.Arrays;

/**
   ListOfRectangle represents a list of Rectangle values that can grow and shrink.
   <p>
   The public interface of this list has the following methods.
   The methods marked with a * are essential methods. The others
   could be left out without the class losing any capability (but
   the users of the class would have to do more work).
   *    int       size()
   *    void      add(int index, Rectangle r)
        void      add(Rectangle r)
   *    Rectangle remove(int index)
        Rectangle remove()
        boolean   remove(Rectangle r)
        void      set(int index, Rectangle r)
   *    Rectangle get(int index)
        void      clear()
        boolean   isEmpty()
        boolean   contains(Rectangle r)
        int       indexOf(Rectangle r)
        int       lastIndexOf(Rectangle r)

http://people.scs.carleton.ca/~lanthier/teaching/COMP1406/Notes/COMP1406_Ch8_AbstractDataTypes.pdf#page=5

https://docs.oracle.com/en/java/javase/11/docs/api/java.base/java/util/List.html
*/
public class ListOfRectangle
{
   private Rectangle[] items; // partially full array holding the Rectangles
   private int itemCount;     // how many Rectangles are in the list

   /**
      Default constructor creates an empty list
      with a backing array of size eight.
   */
   public ListOfRectangle()
   {
      items = new Rectangle[8];
      itemCount = 0;
   }


   /**
      Constructor that creates an empty list
      with a backing array of the given {@code capacity}.

      @param capacity  size for the initial backing array
   */
   public ListOfRectangle(int capacity)
   {
      items = new Rectangle[capacity];
      itemCount = 0;
   }


   /**
      Returns the number of items currently in this {@code List}.

      @return the number of items in this {@code List}
   */
   public int size()
   {
      return itemCount;
   }


   /**
      Returns true if this {@code List} does not conatin any items.

      @return true if there are no items in this {@code List}
   */
   public boolean isEmpty()
   {
      return 0 == itemCount;
    //return 0 == size();   // self encapsulation
   }


   /**
      Return the item at the given {@code index} in this {@code List}.

      @param index  location in this {@code List} of the item to retrieve
      @return a reference to the item in this {@code List} at location {@code index}
   */
   public Rectangle get(int index)
   {
    //if ( index < 0 || index >= size() )   // self encapsulation
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      return items[index];
   }


   /**
      Set the value of the {@code List} item at the given {@code index}.

      @param index  location in this {@code List} of the item to replace
      @param item   reference to the item to place in this {@code List}
   */
   public void set(int index, Rectangle item)
   {
    //if ( index < 0 || index >= size() )   // self encapsulation
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      items[index] = item;
   }


   /**
      Add a new item to the end of this {@code List}.
      The {@code List} size increases by one.

      @param item  reference to the item to be added to this {@code List}
   */
   public void add(Rectangle item)
   {
      if (itemCount == items.length)
      {
         items = Arrays.copyOf( items,    2*items.length ); // good for small arrays
       //items = Arrays.copyOf( items, 1000+items.length ); // good for large arrays
       //items = Arrays.copyOf( items,    1+items.length ); // bad idea
      }

      items[itemCount] = item;
      ++itemCount;
   }


   /**
      Insert the item at the given {@code index} in this {@code List}.
      The {@code List} size increases by one.
      <p>
      For this method, the index is valid if 0 <= index <= size().
      <p>
      If the index is equal to the size of the list, that means that we are
      adding an elelment to the end of the list. In particular, if the list
      is empty, we can add an item at index 0, even though 0 is not (yet)
      a valid index into the list.
      <p>
      Items following the added item are shifted up one space in the backing array.

      @param index  location in this {@code List} for the inserted {@code item}
      @param item   reference to the item to be added to this {@code List}
   */
   public void add(int index, Rectangle item)
   {
      if ( index < 0 || index > itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      if (itemCount == items.length)
      {
         items = Arrays.copyOf( items,    2*items.length ); // good for small arrays
       //items = Arrays.copyOf( items, 1000+items.length ); // good for large arrays
       //items = Arrays.copyOf( items,    1+items.length ); // bad idea
      }

      for (int i = itemCount; i > index; --i)
      {
         items[i] = items[i - 1];
      }

      items[index] = item;
      ++itemCount;
   }


   /**
      Remove the item at the given {@code index} in this {@code List}.
      The {@code List} size decreases by one.

      Items following the removed item are moved down one space in the backing array.

      @param index  location in this {@code List} of the item to remove
      @return a reference to the item that was removed
   */
   public Rectangle remove(int index)
   {
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      Rectangle removedItem = items[index];

      for (int i = index + 1; i < itemCount; ++i)
      {
         items[i - 1] = items[i];
      }

      --itemCount;

      return removedItem;
   }


   /**
      Remove the last item in this {@code List}.
      The {@code List} size decreases by one.

      @return a reference to the item that was removed
   */
   public Rectangle remove()
   {
      --itemCount;

      return items[itemCount];
   }


   /**
      Remove the first occurrence of the given item from this {@code List}.
      Return {@code true} if the {@code item} is found, otherwise return {@code false}.

      Items following the removed item are moved down one space in the backing array.

      @param item  reference to the item to search for and remove from this {@code List}
      @return true if the {@code item} was found and removed, false otherwise
   */
   public boolean remove(Rectangle item)
   {
      int index = this.indexOf(item);

      if (-1 == index)
      {
         return false;
      }

      this.remove(index);

      return true;
   }


   /**
      Remove all the items in this {@code List}.
      The {@code List} size becomes zero.
   */
   public void clear()
   {
      for (int i = 0; i < itemCount; ++i)
      {
         items[i] = null;
      }

      itemCount = 0;
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return {@code true} if the {@code item} is found.

      @param item  reference to the item to search for in this {@code List}
      @return true if the {@code item} was found in this {@code List}
   */
   public boolean contains(Rectangle item)
   {
      return -1 != this.indexOf(item);
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the first occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the first index of {@code item} if it is found, -1 otherwise
   */
   public int indexOf(Rectangle item)
   {
      for (int i = 0; i < itemCount; ++i)
      {
         if ( item == null ? item == items[i] : item.equals(items[i]) )
         {
            return i;
         }
      }

      return -1;
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the last occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the last index of {@code item} if it is found, -1 otherwise
   */
   public int lastIndexOf(Rectangle item)
   {
      for (int i = itemCount - 1; i >= 0; --i)
      {
         if ( item == null ? item == items[i] : item.equals(items[i]) )
         {
            return i;
         }
      }

      return -1;
   }


   /**
      Return a {@link String} representation of this {@code List}.

      @return {@link String} representation of this {@code List}
   */
   @Override
   public String toString()
   {
      String result = "";
/*
      for (int i = 0; i < this.size(); ++i) // self encapsulated version
      {
         result += this.get(i) + "\n";
      }
*/
      for (int i = 0; i < itemCount; ++i)
      {
         result += items[i] + "\n";
      }

      return result;
   }
}
